import torch
from torch import nn
from transformers import RobertaPreTrainedModel, RobertaModel, PreTrainedModel


class RobertaClassificationHeadPruned(nn.Module):
    """Head for sentence-level classification tasks."""

    def __init__(self, config):
        super().__init__()
        self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        classifier_dropout = (
            config.classifier_dropout if config.classifier_dropout is not None else config.hidden_dropout_prob
        )
        self.dropout = nn.Dropout(classifier_dropout)
        self.out_proj_adjusted = nn.Linear(config.hidden_size*2, 1)
        
    def forward_direction(self, features):
        x = features[:, 0, :]  # take <s> token (equiv. to [CLS])
        x = self.dropout(x)
        x = self.dense(x)
        x = torch.tanh(x)
        x = self.dropout(x)
        return x

    def forward(self, features_forward, features_backward, **kwargs):
        
        x_forward = self.forward_direction(features_forward)
        x_backward = self.forward_direction(features_backward)
        
        concatenated_hs = torch.cat([x_forward, x_backward], -1)
                
        x = self.out_proj_adjusted(concatenated_hs)
        return x


class TwoFoldRoberta(RobertaPreTrainedModel):
    _keys_to_ignore_on_load_missing = [r"position_ids"]

    def __init__(self, config):
        super().__init__(config)
        self.num_labels = config.num_labels
        self.config = config

        self.roberta = RobertaModel(config, add_pooling_layer=False)
        self.classifier = RobertaClassificationHeadPruned(config)

        # Initialize weights and apply final processing
        self.post_init()

    def forward(
        self,
        batch_forward, batch_backward,
    ):
        r"""
        labels (`torch.LongTensor` of shape `(batch_size,)`, *optional*):
            Labels for computing the sequence classification/regression loss. Indices should be in `[0, ...,
            config.num_labels - 1]`. If `config.num_labels == 1` a regression loss is computed (Mean-Square loss), If
            `config.num_labels > 1` a classification loss is computed (Cross-Entropy).
        """

        outputs_forward = self.roberta(
            **batch_forward
        )[0]
        
        outputs_backward = self.roberta(
            **batch_backward
        )[0]
                
        logits = self.classifier(outputs_forward, outputs_backward)
                
        return logits



from typing import List

import torch
from transformers import AutoTokenizer
from torch.utils.data import Dataset, DataLoader
from tqdm.auto import tqdm

from .mis_architecture import TwoFoldRoberta


class PairsDatasetInference(Dataset):
    def __init__(self, texts_first, texts_second):
        self.texts_first = texts_first
        self.texts_second = texts_second

    def __len__(self):
        return len(self.texts_first)

    def __getitem__(self, idx):
        return self.texts_first[idx], self.texts_second[idx]


class MIS:
    def __init__(
        self,
        model_name='SkolkovoInstitute/Mutual_Implication_Score',
        device=None
    ):
        """
            model_name: name or path to the pretrained model (by default, 'SkolkovoInstitute/Mutual_Implication_Score')
            device: name of the pytorch device (by default, 'cuda' or 'cpu', if CUDA is not available)
        """
        self.model_name = model_name
        self.device = torch.device(device or 'cuda' if torch.cuda.is_available() else 'cpu')
        self.model = TwoFoldRoberta.from_pretrained(model_name).to(self.device)
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)

    def compute(self, source_texts, paraphrases, verbose=True, batch_size=16) -> List[float]:
        dataset_direct = PairsDatasetInference(source_texts, paraphrases)
        dataloader_direct = DataLoader(dataset_direct, batch_size=batch_size)

        dataset_reverse = PairsDatasetInference(paraphrases, source_texts)
        dataloader_reverse = DataLoader(dataset_reverse, batch_size=batch_size)

        iterator = zip(dataloader_direct, dataloader_reverse)
        if verbose:
            iterator = tqdm(iterator, total=len(dataloader_reverse))

        preds = []

        for b1, b2 in iterator:
            with torch.no_grad():
                tokenized1 = self.tokenizer(*b1, padding=True, truncation='longest_first',
                                            return_tensors="pt").to(self.device)
                tokenized2 = self.tokenizer(*b2, padding=True, truncation='longest_first',
                                            return_tensors="pt").to(self.device)
                merged_prob = self.model(tokenized1, tokenized2)
                merged_prob = torch.sigmoid(merged_prob)

            merged_prob = merged_prob.cpu().numpy()
            preds.extend(merged_prob)

        preds_float = [float(e) for e in preds]
        return preds_float



import os
import setuptools

description = "Mutual implication score: a symmetric measure of text semantic similarity " \
              "based on a RoBERTA model pretrained for natural language inference."
long_description = description
if os.path.exists("README.md"):
    with open("README.md", "r", encoding="utf-8") as fh:
        long_description = fh.read()


setuptools.setup(
    name="mutual_implication_score",
    version="0.0.0",
    author="Nikolai Babakov and David Dale",
    author_email="bbk_junior@mail.ru",
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/skoltech-nlp/mutual_implication_score",
    packages=setuptools.find_packages(),
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires=['transformers>=4.13.0', 'torch', 'tqdm'],
)
